from lupin import Mapper, Schema, fields as f, validators as v
from lupin.errors import InvalidDocument, InvalidLength
from models import Artist

mapper = Mapper()

artist_schema = Schema({
    "name": f.String(validators=v.Length(max=10)),
})

mapping = mapper.register(Artist, artist_schema)

data = {
    "name": "Leonardo da Vinci"
}

try:
    mapper.load(data, artist_schema, allow_partial=True)
except InvalidDocument as errors:
    error = errors[0]
    assert isinstance(error, InvalidLength)
    assert error.path == ["name"]



from datetime import datetime
import pytest

from lupin.fields import List, DateTime
from lupin.errors import InvalidType
from lupin import Mapper


@pytest.fixture
def mapper():
    return Mapper()


@pytest.fixture
def field():
    return List(DateTime(format="%Y-%m-%d"))


@pytest.fixture
def datetimes():
    return [
        datetime(2017, 1, 1),
        datetime(2018, 1, 1)
    ]


@pytest.fixture
def datetime_strings():
    return [
        "2017-01-01",
        "2018-01-01"
    ]


class TestLoad(object):
    def test_returns_a_list_of_datetime(self, field, datetimes, datetime_strings, mapper):
        result = field.load(datetime_strings, mapper)
        assert result == datetimes

    def test_returns_none_if_value_is_null(self, field, mapper):
        assert field.load(None, mapper) is None


class TestDump(object):
    def test_returns_a_list_of_datetime_strings(self, field, datetimes, datetime_strings, mapper):
        result = field.dump(datetimes, mapper)
        assert result == datetime_strings

    def test_returns_none_if_value_is_null(self, field, mapper):
        assert field.dump(None, mapper) is None


class TestValidate(object):
    def test_raise_exception_if_not_list(self, field):
        with pytest.raises(InvalidType):
            field.validate("", [], mapper)

    def test_raise_exception_if_an_item_is_invalid(self, field, mapper):
        with pytest.raises(InvalidType) as exc:
            field.validate([46], [], mapper)

        error = exc.value
        assert error.path == ["0"]

    def test_does_nothing_if_all_items_are_valid(self, field, datetime_strings, mapper):
        field.validate(datetime_strings, [], mapper)



# coding: utf-8
import pytest

from lupin import Mapper, Mapping, Schema, String
from lupin.errors import MissingMapping, SchemaAlreadyRegistered, InvalidDocument
from tests.fixtures import Thief


@pytest.fixture
def mapper():
    return Mapper()




class TestRegister(object):
    def test_sets_the_default_schema(self, mapper, thief_schema, thief):
        mapper.register(Thief, thief_schema)
        data = mapper.dump(thief)
        assert data == {
            "firstName": "Arsène",
            "lastName": "Lupin"
        }

    def test_raise_error_if_schema_already_registered(self, mapper, thief_schema):
        mapper.register(Thief, thief_schema)

        with pytest.raises(SchemaAlreadyRegistered):
            mapper.register(Thief, thief_schema)


class TestDump(object):
    @pytest.fixture(autouse=True)
    def mapping(self, mapper, thief_schema):
        mapper.register(Thief, thief_schema)

    def test_uses_default_mapping(self, mapper, thief, thief_data):
        data = mapper.dump(thief)
        assert data == thief_data

    def test_uses_schema(self, mapper, thief, thief_data, thief_schema):
        data = mapper.dump(thief, thief_schema)
        assert data == thief_data

    def test_dumps_list_of_objects(self, mapper, thief, thief_data):
        data = mapper.dump([thief])
        assert data == [thief_data]

    def test_raises_exception_if_no_mapping(self, mapper):
        with pytest.raises(MissingMapping):
            mapper.dump(46)


class TestLoad(object):
    @pytest.fixture(autouse=True)
    def mapping(self, mapper, thief_schema):
        mapper.register(Thief, thief_schema)

    def test_returns_a_thief(self, mapper, thief_data, thief_schema):
        thief = mapper.load(thief_data, thief_schema)
        assert isinstance(thief, Thief)

    def test_loads_list(self, mapper, thief_data, thief_schema):
        thieves = mapper.load([thief_data], thief_schema)
        assert isinstance(thieves, list)
        assert isinstance(thieves[0], Thief)


class TestValidate(object):
    @pytest.fixture(autouse=True)
    def mapping(self, mapper, thief_schema):
        mapper.register(Thief, thief_schema)

    def test_does_nothing_if_valid(self, mapper, thief_data, thief_schema):
        mapper.validate(thief_data, thief_schema)

    def test_validates_a_valid_list(self, mapper, thief_data, thief_schema):
        mapper.validate([thief_data], thief_schema)

    def test_raise_error_if_invalid(self, mapper, thief_data, thief_schema):
        with pytest.raises(InvalidDocument):
            mapper.validate({}, thief_schema)


class TestLoadAttrs(object):
    @pytest.fixture(autouse=True)
    def mapping(self, mapper, thief_schema):
        mapper.register(Thief, thief_schema)

    def test_returns_a_dict(self, mapper, thief_data, thief_schema):
        attrs = mapper.load_attrs(thief_data, thief_schema)
        assert attrs == {
            "first_name": "Arsène",
            "last_name": "Lupin"
        }

    def test_loads_list(self, mapper, thief_data, thief_schema):
        thieves = mapper.load_attrs([thief_data], thief_schema)
        assert isinstance(thieves, list)
        assert isinstance(thieves[0], dict)


class TestGetObjectMapping(object):
    @pytest.fixture(autouse=True)
    def mapping(self, mapper, thief_schema):
        mapper.register(Thief, thief_schema)

    def test_returns_mapping_of_object(self, mapper, thief):
        mapping = mapper.get_object_mapping(thief)
        assert mapping.cls == type(thief)

    def test_raises_exception_if_no_mapping(self, mapper):
        with pytest.raises(MissingMapping):
            mapper.get_object_mapping(46)

    def test_restrict_access_to_some_schemas(self, mapper, thief, thief_schema):
        schema = Schema({})
        mapper.register(Thief, schema)
        mapping = mapper.get_object_mapping(thief, schemas=[thief_schema])
        assert mapping.schema == thief_schema

    def test_raise_error_if_no_mapping_among_schemas_provided(self, mapper):
        with pytest.raises(MissingMapping):
            mapper.get_object_mapping(46, schemas=[])

    def test_returns_a_mapping_from_the_super_class(self, mapper, thief_schema):
        class JuniorThief(Thief):
            pass

        thief = JuniorThief("John", "Doe")
        mapping = mapper.get_object_mapping(thief)
        data = mapping.dump(thief, mapper)
        assert data == {
            "firstName": "John",
            "lastName": "Doe"
        }



import datetime
from lupin import fields, Mapper
import pytest


@pytest.fixture
def mapper():
    return Mapper()


@pytest.fixture
def field():
    return fields.Date()


@pytest.fixture
def date():
    return datetime.date(2017, 10, 7)


class TestDump(object):
    def test_returns_date_as_string(self, field, date, mapper):
        assert field.dump(date, mapper) == "2017-10-07"

    def test_returns_none_if_value_is_null(self, field, mapper):
        assert field.dump(None, mapper) is None


class TestLoad(object):
    def test_returns_a_date(self, field, date, mapper):
        assert field.load("2017-10-07", mapper) == date

    def test_returns_none_if_value_is_null(self, field, mapper):
        assert field.load(None, mapper) is None



from datetime import datetime

from . import Validator
from ..errors import InvalidDateTimeFormat


class DateTimeFormat(Validator):
    """Validates date and datetimes format"""

    def __init__(self, format):
        """
        Args:
            format (str): datetime format
        """
        self._format = format

    def __call__(self, value, path):
        """Validate that value can be parsed as a date

        Args:
            value (str): datetime string to validate
            path (list): error path
        """
        try:
            datetime.strptime(value, self._format)
        except ValueError:
            raise InvalidDateTimeFormat(value, self._format, path)



import pytest

from lupin.validators import Equal
from lupin.errors import NotEqual


@pytest.fixture
def validator():
    return Equal(46)


class TestCall(object):
    def test_raise_error_if_invalid_value(self, validator):
        with pytest.raises(NotEqual):
            validator("46", [])

    def test_does_nothing_if_valid(self, validator):
        validator(46, [])



from datetime import datetime

from . import Field
from .compatibility import merge_validator
from ..validators import Type


class DateTime(Field):
    """Field used to handle datetime values"""

    def __init__(self, format="%Y-%m-%dT%H:%M:%S", *args, **kwargs):
        """
        Args:
            format (str): datetime format to use
        """
        merge_validator(kwargs, Type(str))
        super(DateTime, self).__init__(*args, **kwargs)
        self._format = format

    def load(self, value, mapper):
        """Loads a datetime python object from a JSON string

        Args:
            value (str): a value
            mapper (Mapper): mapper used to load data

        Returns:
            datetime
        """
        if value is None:
            return None

        return datetime.strptime(value, self._format)

    def dump(self, value, mapper):
        """Dump a datetime to string representation

        Args:
            value (datetime): a datetime
            mapper (Mapper): mapper used to dump data

        Returns:
            str
        """
        if value is None:
            return None

        return value.strftime(self._format)



from lupin.fields.compatibility import merge_validator
from lupin.validators import Equal
from lupin.errors import ValidationError
import pytest


@pytest.fixture
def invalid():
    return Equal("sernine")


@pytest.fixture
def valid():
    return Equal("lupin")


class TestMergeValidator(object):
    def test_add_validator_to_list(self, valid, invalid):
        kwargs = {"validators": [valid]}
        merge_validator(kwargs , invalid)
        with pytest.raises(ValidationError):
            kwargs["validators"]("lupin", [])

    def test_add_validator_to_validator(self, valid, invalid):
        kwargs = {"validators": valid}
        merge_validator(kwargs , invalid)
        with pytest.raises(ValidationError):
            kwargs["validators"]("lupin", [])

    def test_set_validator_if_no_validators(self, invalid):
        kwargs = {}
        merge_validator(kwargs , invalid)
        assert kwargs["validators"] is invalid



# -*- coding: utf-8 -*-
from lupin import ValidatorsOrCombination, ValidatorsAndCombination
from lupin.errors import ValidationError
from lupin.validators import Equal
import pytest


@pytest.fixture
def invalid():
    return Equal("sernine")


@pytest.fixture
def valid():
    return Equal("lupin")


class TestCall(object):
    def test_raise_error_if_all_validators_are_invalid(self, invalid):
        combination = ValidatorsOrCombination([invalid, invalid, invalid])
        with pytest.raises(ValidationError):
            combination("andrésy", [])

    def test_do_not_raise_error_if_one_validator_is_valid(self, invalid, valid):
        combination = ValidatorsOrCombination([invalid, valid, invalid])
        combination("lupin", [])

    def test_do_not_raise_error_if_complex_expression_is_valid(self, invalid, valid):
        combination = (valid & valid) | invalid & valid & valid | invalid
        combination("lupin", [])

    def test_raise_error_if_complex_expression_is_invalid(self, invalid, valid):
        combination = (valid & valid) | invalid & valid & valid | invalid
        with pytest.raises(ValidationError):
            combination("andrésy", [])


class TestOr(object):
    def test_returns_self_if_other_is_a_validator(self, valid):
        combination = valid | valid
        assert (combination | valid) is combination

    def test_returns_self_if_other_is_or_combination(self, valid):
        combination = valid | valid
        combination2 = valid | valid
        assert (combination | combination2) is combination

    def test_else_returns_a_new_combination(self, valid):
        combination = valid | valid
        combination2 = valid & valid
        result = combination | combination2
        assert result is not combination
        assert isinstance(result, ValidatorsOrCombination)


class TestAnd(object):
    def test_returns_an_and_combination(self, valid):
        combination = valid | valid
        result = combination & valid
        assert isinstance(result, ValidatorsAndCombination)



# -*- coding: utf-8 -*-
import pkg_resources

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except:
    __version__ = 'unknown'

from .object_factories import constructor, bind  # NOQA
from .validators_combination import ValidatorsOrCombination, ValidatorsAndCombination, ValidatorsNullCombination  # NOQA
from .fields import *  # NOQA
from .schema import Schema  # NOQA
from .mapping import Mapping  # NOQA
from .mapper import Mapper  # NOQA



import re

import pytest

from lupin.errors import InvalidMatch
from lupin.validators import Match


@pytest.fixture
def validator():
    regexp = re.compile("hello")
    return Match(regexp)


class TestCall(object):
    def test_raise_error_if_does_not_match(self, validator):
        with pytest.raises(InvalidMatch) as exc:
            validator("bye", [])

    def test_does_nothing_if_match(self, validator):
        validator("hello", [])

    def test_raise_error_value_is_not_a_string(self, validator):
        with pytest.raises(InvalidMatch) as exc:
            validator(None, [])



from . import Validator
from ..errors import NotNone


class IsNone(Validator):
    """Validate that a value is None"""

    def __call__(self, value, path):
        """Validate that value is None

        Args:
            value (object): value to validate
            path (list): error path
        """
        if value is not None:
            raise NotNone(value, path)



from datetime import datetime
from lupin import Mapper, Schema, fields as f
from models import Artist, Diamond, Painting, Thief

leonardo = Artist(name="Leonardo da Vinci", birth_date=datetime(1452, 4, 15))
mona_lisa = Painting(name="Mona Lisa", author=leonardo)
arsene = Thief(name="Arsène Lupin", stolen_items=[mona_lisa])


mapper = Mapper()
artist_schema = Schema({
    "name": f.String(),
    "birthDate": f.DateTime(binding="birth_date", format="%Y-%m-%d")
})

diamond_schema = Schema({
    "carat": f.Field(),
    "type": f.Constant("diamond")  # this will be used to know which mapping to used while loading JSON
})

painting_schema = Schema({
    "name": f.String(),
    "type": f.Constant("painting"),
    "author": f.Object(artist_schema)
})

# Use `PolymorphicList` for `stolen_items`
thief_schema = Schema({
    "name": f.String(),
    "stolenItems": f.PolymorphicList(on="type",  # JSON key to lookup for the polymorphic type
                                     binding="stolen_items",
                                     schemas={
                                         "painting": painting_schema,  # if `type == "painting"` then use painting_schema
                                         "diamond": diamond_schema  # if `type == "diamond"` then use diamond_schema
                                     })
})


mapper.register(Artist, artist_schema)
mapper.register(Diamond, diamond_schema)
mapper.register(Painting, painting_schema)
mapper.register(Thief, thief_schema)


diamond = Diamond(carat=20)
arsene.stolen_items.append(diamond)

# Dump object
data = mapper.dump(arsene)
assert data == {
    "name": "Arsène Lupin",
    "stolenItems": [
        {
            "name": "Mona Lisa",
            "type": "painting",
            "author": {
                "name": "Leonardo da Vinci",
                "birthDate": "1452-04-15"
            }
        },
        {
            "carat": 20,
            "type": "diamond"
        }
    ]
}

# Load data
thief = mapper.load(data, thief_schema)
assert isinstance(thief.stolen_items[0], Painting)
assert isinstance(thief.stolen_items[1], Diamond)



import pytest

from lupin.fields import Int
from lupin.errors import InvalidType
from lupin import Mapper

@pytest.fixture
def mapper():
    return Mapper()


@pytest.fixture
def field():
    return Int()


class TestValidate(object):
    def test_raises_error_if_not_an_int(self, field, mapper):
        with pytest.raises(InvalidType):
            field.validate("46", [], mapper)

    def test_does_nothing_if_int(self, field, mapper):
        field.validate(46, [], mapper)



from . import Field
from .compatibility import merge_validator
from ..validators import Type


class Object(Field):
    """Field used for nested objects.
    It uses a mapping in order to know what class to load and the schema
    used for the JSON representation.
    """

    def __init__(self, schema, *args, **kwargs):
        """
        Args:
            schema (Schema): schema of nested object
        """
        merge_validator(kwargs, Type(dict))
        super(Object, self).__init__(*args, **kwargs)
        self._schema = schema

    def load(self, value, mapper):
        """Loads python object from JSON object

        Args:
            value (dict): nested JSON object
            mapper (Mapper): mapper used to load data

        Returns:
            object
        """
        if value is None:
            return None

        return mapper.load(value, self._schema)

    def dump(self, value, mapper):
        """Dump object into its JSON representation

        Args:
            value (object): an object
            mapper (Mapper): mapper used to dump data

        Returns:
            dict
        """
        if value is None:
            return value

        return self._schema.dump(value, mapper)

    def validate(self, value, path, mapper):
        """Validate value against mapping validators.

        Args:
            value (list): value to validate
            path (list): JSON path of value
            mapper (Mapper): mapper used to dump data
        """
        super(Object, self).validate(value, path, mapper)
        self._schema.validate(value, mapper, path=path)



import lupin


class Validator(object):
    """Validators base class"""

    def __and__(self, other):
        """Combine current validator with another one using the & operator

        Returns:
            ValidatorsAndCombination
        """
        return lupin.ValidatorsAndCombination([self, other])

    def __or__(self, other):
        """Combine current validator with another one using the | operator

        Returns:
            ValidatorsOrCombination
        """
        return lupin.ValidatorsOrCombination([self, other])



from . import Field
from .compatibility import merge_validator
from ..validators import Type


class List(Field):
    """Field handling lists containing same object types.
    It uses an inner field to handle items mapping.
    """

    def __init__(self, field, **kwargs):
        """
        Args:
            field (Field): a field handling list items
        """
        merge_validator(kwargs, Type(list))
        super(List, self).__init__(**kwargs)
        self._field = field

    def load(self, value, mapper):
        """Loads list of python objects from JSON list

        Args:
            value (list): JSON list
            mapper (Mapper): mapper used to load data

        Returns:
            list
        """
        if value is None:
            return None

        return [self._field.load(item, mapper) for item in value]

    def dump(self, value, mapper):
        """Dump list of object

        Args:
            value (list): list of
            mapper (Mapper): mapper used to dump data

        Returns:
            list
        """
        if value is None:
            return None

        return [self._field.dump(item, mapper) for item in value]

    def validate(self, value, path, mapper):
        """Validate each items of list against nested field validators.

        Args:
            value (list): value to validate
            path (list): JSON path of value
            mapper (Mapper): mapper used to validate data
        """
        super(List, self).validate(value, path, mapper)
        for item_index, item in enumerate(value):
            item_path = path + [str(item_index)]
            self._field.validate(item, item_path, mapper)
